
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>

#include "video/gbuffer.h"
#include "display/display.h"
#include "cpu/65el02.h"
#include "memory/memory.h"
#include "drive/drive.h"
#include "monitor.h"
#include "screen.h"
#include "options.h"
#include "filein.h"

static char hdr[] = "File Input";
static int cpu_ticks;

void filein_ungetc(int ch)
{
  if ((filein != NULL) && (filein->stream != NULL))
  {
    ungetc(ch, filein->stream);
  }
}

int filein_getc()
{
  int ch = EOF;
  if ((cpu_ticks < cpu.ticks) && (filein != NULL) && (filein->stream != NULL))
  {
    cpu_ticks = cpu.ticks;
    ch = getc(filein->stream);
    if (ch == EOF)
    {
      fclose(filein->stream);
      filein->stream = NULL;
    }
  }
  return ch;
}

void filein_init()
{
  filein = (struct filein_struct *) malloc(sizeof(struct filein_struct));
  memset(filein, 0, sizeof(struct filein_struct));
  filein->filename = (char *) malloc(MAX_INPUT_FILENAME);
  memset(filein->filename, 0, MAX_INPUT_FILENAME);
  cpu_ticks = cpu.ticks;
}

void filein_exit()
{
  if (filein->stream != NULL)
  {
    fclose(filein->stream);
    filein->stream = NULL;
  }
  free(filein->filename);
  free(filein);
  filein = NULL;
}

static int input_abort()
{
  memset(filein->filename, 0, MAX_INPUT_FILENAME);
  filein->length = 0;
  filein->stream = NULL;
  return 1;
}

static int input_okay()
{
  filein->filename[filein->length] = '\0';
  ++filein->length;
  filein->stream = fopen(filein->filename, "r");
  cpu_ticks = cpu.ticks;
  return 1;
}

static void input_append(char ch)
{
  if (filein->length < (MAX_INPUT_FILENAME-1))
  {
    scrprintf(SCR_ID_INPUT, "%c", ch);
    filein->filename[filein->length] = ch;
    ++(filein->length);
  }
}

static void input_clear()
{
	scrclr(SCR_ID_INPUT);
	scrattr(SCR_ID_INPUT, 0x7b);
	scrprintf(SCR_ID_INPUT, "%*s", SCR_COLS, "");
	scrxy(SCR_ID_INPUT, (80-strlen(hdr))/2, 0);
	scrprintf(SCR_ID_INPUT, "%s", hdr);
	scrattr(SCR_ID_INPUT, 0xbe);
	scrxy(SCR_ID_INPUT,0,2);
	scrprintf(SCR_ID_INPUT,"Please enter filename:\n ");
  filein->length = 0;
  if (filein->stream != NULL)
  {
    fclose(filein->stream);
    filein->stream = NULL;
  }
}

int optfilein()
{
	int esc = 0, quit = 0, last_tick = -1;

	openscreen(SCR_ID_INPUT);
	scrblink(SCR_ID_INPUT, 0);
	drawstring(0, C_SIZEY-1, "\r ESC - Back  RETURN - Okay  BACKSPACE - Clear", 0x0f, 0x09, rom_charset);
  input_clear();
  
  if (filein->stream != NULL)
  {
    fclose(filein->stream);
    filein->stream = NULL;
  }
  
	while (!quit && !esc)
	{
		displayscreen(SCR_ID_INPUT);
		updateptc();

		while (last_tick == timer_tick)
			SDL_Delay(1);

		last_tick = timer_tick;

		while (getkeys())
		{
			if (lastkeysym >= 0)
			{
				if (lastkeysym == SDLK_ESCAPE)
					esc = input_abort();
        
        else if (lastkeysym == SDLK_RETURN)
          esc = input_okay();
        
				else if ((lastkeysym == SDLK_x) && (lastkeymod & KMOD_ALT))
					quit = 1;	// alt-x
        
				else if ((lastkeysym == SDLK_F4) && (lastkeymod & KMOD_ALT))
					quit = 1;	// alt-f4

        else if (lastkeysym == SDLK_BACKSPACE)
          input_clear();

			}

			if ((lastkey > 0) && !quit && !esc)
        input_append((char)lastkey);

			if (lastkey == -2)
				quit = 1;
    }
  }
  
  return quit;
}
